import os
import subprocess
import difflib
from termcolor import colored


class MissingFilePathError(ValueError):
    def __str__(self):
        return "A file path is required when creating a FileEditor instance."


class FileEditor:
    CONFIRMATION_REQUEST = "Do you want to apply the changes? (y/n): "
    CONFIRMATION_CHARACTER = "y"
    CHANGES_APPLIED_TEXT = "Changes applied to: "
    CHANGES_DISCARDED_TEXT = "Changes discarded."

    DEFAULT_EDITOR = "vim"

    def __init__(self, file_path=None):
        if file_path is None:
            raise MissingFilePathError
        self.file_path = file_path

    def exists(self):
        """
        Returns True if the file exists, False otherwise.
        """
        return os.path.exists(self.file_path)

    def content(self):
        """
        Reads the file and returns the content as a string.
        """
        with open(self.file_path, "r") as file:
            content = file.read()
        return content

    def open_editor(self):
        """
        Opens the file in the default editor
        """
        editor = os.environ.get("EDITOR", self.DEFAULT_EDITOR)
        subprocess.call([editor, self.file_path])

    def apply_changes(self, content):
        """
        Takes a file and content string as inputs. It generates a diff
        comparing the string to the contents of the file. It then displays
        the diff and asks the user to confirm the changes. If they confirm,
        it writes the content to the file.

        :param content: The content to be written to the file
        """
        diff = self._calculate_diff(content)
        confirmed = self._get_confirmation(diff)
        if confirmed:
            self._write(content)
            print(f"{self.CHANGES_APPLIED_TEXT}{self.file_path}")
            return True
        print(self.CHANGES_DISCARDED_TEXT)
        return False

    def _write(self, content):
        with open(self.file_path, "w") as file:
            file.write(content)

    def _get_confirmation(self, diff):
        self._print_diff(diff)
        response = input(self.CONFIRMATION_REQUEST)
        if response.lower() == self.CONFIRMATION_CHARACTER:
            return True
        return False

    def _calculate_diff(self, content):
        if self.exists():
            file_content = self.content()
        else:
            file_content = ""
        return difflib.unified_diff(
            file_content.splitlines(), content.splitlines(), lineterm=""
        )

    def _print_diff(self, diff):
        for line in diff:
            self._print_colored_line(line)

    def _print_colored_line(self, line):
        if line.startswith("+"):
            print(colored(line, "green"))
        elif line.startswith("-"):
            print(colored(line, "red"))
        else:
            print(line)



from abc import ABC, abstractmethod

DEFAULT_SYSTEM_MESSAGE = "You are a helpful assistant."
CODE_SYSTEM_MESSAGE = """You are a code generating assistant.
You obey the following rules:
- You only respond only in code.
- You respond only in complete files.
- Any explanation should be included only as inline comments.
- Always wrap your code in triple backticks.
- You do not include usage examples."""


class LanguageModel(ABC):
    """
    Abstract base class representing a large language model. Delegates
    chat response logic out to implementations in model specific subclasses.

    Concrete language model classes must implement _respond_from_prompt to
    take a prompt string as a param and return a string as an answer.
    """

    def __init__(self, model_name: str):
        self.model_name = model_name

    def respond(self, prompt: str) -> str:
        """
        Respond to a prompt with a general instruction message prepended.

        :param prompt: The prompt to respond to.
        :return: The response string.
        """
        return self._respond_from_prompt(prompt, DEFAULT_SYSTEM_MESSAGE)

    def respond_with_code(self, prompt: str) -> str:
        """
        Respond to a prompt with a code generation instruction message prepended.

        :param prompt: The prompt to respond to.
        :return: The response string.
        """
        return self._respond_from_prompt(prompt, CODE_SYSTEM_MESSAGE)

    @abstractmethod
    def _respond_from_prompt(self, prompt: str, system_message: str) -> str:
        pass



import os
import configparser
from pathlib import Path

USER_CONFIG_FILE = Path.home() / ".nailrc"


def get_api_key():
    if "OPENAI_API_KEY" in os.environ:
        return os.environ["OPENAI_API_KEY"]

    if USER_CONFIG_FILE.is_file():
        config = configparser.ConfigParser()
        config.read(USER_CONFIG_FILE)
        if "openai" in config and "api_key" in config["openai"]:
            return config["openai"]["api_key"]

    return None


def save_api_key(api_key):
    config = configparser.ConfigParser()
    if USER_CONFIG_FILE.is_file():
        config.read(USER_CONFIG_FILE)

    if "openai" not in config:
        config["openai"] = {}

    config["openai"]["api_key"] = api_key

    with USER_CONFIG_FILE.open("w") as config_file:
        config.write(config_file)



import os
import re

from pathlib import Path
from typing import List
from nail.core.prompt.formatting_utils import file_block


class ContextCompiler:
    """
    Compiles prompt context from the files in the given context_file_paths.

    :param context_file_paths: A list of file paths to include in the context.
    :param ignore_list: A list of file names or regex patterns to ignore.
    """

    CONTEXT_PREFIX = "Existing files for context:"
    # TODO: Make this list configurable
    DEFAULT_IGNORE_LIST = ["README", "LICENSE", "^[._]", "^test", "test$"]

    def __init__(
        self, context_file_paths=[os.getcwd()], ignore_list=DEFAULT_IGNORE_LIST
    ):
        self.context_file_paths = context_file_paths
        self.ignore_list = ignore_list

    def compile_all(self):
        """
        Compiles prompt context from all files in the given context_file_paths.
        This includes a prefix explaining the context, and a code block
        and file name label for each file.

        :return: A string containing the prompt context.
        """
        all_files = self._list_all_files()
        return self._compile_context(all_files)

    def compile_all_minus_ignored(self):
        """
        Compiles prompt context from given context_file_paths. Includes all
        files in the given paths, minus any that are included in the
        ContextCompiler's ignore_list. Context includes a prefix explaining the
        context, and a code block and file name label for each file.

        :return: A string containing the prompt context.
        """
        relevant_files = self._filter_ignored(self._list_all_files())
        return self._compile_context(relevant_files)

    def _compile_context(self, files):
        context = [file_block(file) for file in files]
        return f"{self.CONTEXT_PREFIX}\n\n{''.join(context)}"

    def _list_all_files(self):
        all_file_paths = []
        for path in self.context_file_paths:
            file_paths = self._list_files_at_path(path)
            all_file_paths.extend(file_paths)
        return all_file_paths

    def _list_files_at_path(self, path):
        if os.path.isfile(path):
            return [path]
        file_paths = []
        for root, dirs, files in os.walk(path):
            dirs[:] = [d for d in dirs if not d.startswith(".")]
            for file in files:
                file_paths.append(os.path.join(root, file))
        return file_paths

    def _filter_ignored(self, file_paths):
        return [
            file_path for file_path in file_paths if not self._is_ignored(file_path)
        ]

    def _is_ignored(self, file_path):
        # Generate regexes for each item in the ignore list and
        # return True if any of them match the given file path
        file = os.path.basename(file_path)
        return any(re.compile(item).search(file) for item in self.ignore_list)



import itertools
import sys
import time

from threading import Event, Thread

ANIMATION_FRAME_SECONDS = 0.04
DEFAULT_LOADING_MESSAGE = "Loading..."

FRAMES = [
    "|==========>                    |",
    " |==========>                   |",
    "  |==========>                  |",
    "   |==========>                 |",
    "    |==========>                |",
    "     |==========>               |",
    "      |==========>              |",
    "       |==========>             |",
    "        |==========>            |",
    "         |==========>           |",
    "          |==========>          |",
    "           |==========>         |",
    "            |==========>        |",
    "             |==========>       |",
    "              |==========>      |",
    "               |==========>     |",
    "                |==========>    |",
    "                 |==========>   |",
    "                  |==========>  |",
    "                   |==========> |",
    "                    |==========>|",
    "                     |==========|",
    "                      |=========|",
    "                       |========|",
    "                        |=======|",
    "                         |======|",
    "                          |=====|",
    "                           |====|",
    "                            |===|",
    "                             |==|",
    "                              |=|",
    "                               ||",
    "                               ||",
    "                               ||",
    "                               ||",
    "                               ||",
    "                               ||",
    ">                               |",
    "=>                              |",
    "==>                             |",
    "===>                            |",
    "====>                           |",
    "=====>                          |",
    "======>                         |",
    "=======>                        |",
    "========>                       |",
    "=========>                      |",
    "==========>                     |",
]


def loadable(func, loading_message):
    """
    This decorator is used to display a loading animation
    while a function is running.

    :param func: The function to be decorated
    :param loading_message: The message to be displayed before the animation
    """

    def wrapper(*args, **kwargs):
        # Start a thread to display the loading animation
        stop_loading = Event()
        loader_thread = Thread(
            target=_loading,
            args=(
                stop_loading,
                loading_message,
            ),
        )
        loader_thread.start()

        try:
            result = func(*args, **kwargs)
        except Exception as e:
            # Stop the loading animation thread if an error occurs
            _clear_loader(stop_loading, loader_thread, loading_message)
            raise e

        _clear_loader(stop_loading, loader_thread, loading_message)

        return result

    return wrapper


def _loading(stop_loading, loading_message):
    while not stop_loading.is_set():
        _display_loader(stop_loading, loading_message)


def _display_loader(stop_loading, prefix=DEFAULT_LOADING_MESSAGE):
    spinner = itertools.cycle(FRAMES)
    while not stop_loading.is_set():
        sys.stdout.write(f"{prefix}{next(spinner)}")
        sys.stdout.flush()
        time.sleep(ANIMATION_FRAME_SECONDS)
        sys.stdout.write("\r")
    sys.stdout.write("\n")


def _clear_loader(stop_loading, loader_thread, loading_message):
    # Stop the loading animation thread
    stop_loading.set()

    # Erase the animation line after the decorated function completes
    sys.stdout.write("\r" + " " * (len(loading_message) + len(FRAMES[-1])) + "\r")
    sys.stdout.flush()

    loader_thread.join()



from nail.core.file_editor import FileEditor
from nail.core.chat import Chat
from nail.core.prompt.prompt import BuildReadmePrompt


def build_readme(readme_file_path, model=None):
    """
    Gathers context from all files in the current directory, builds a prompt for
    OpenAI to generate a README file for the application, calls the predict_readme
    method, and writes the generated file to the specified path.

    :param readme_file_path: Path to save the generated README file
    """
    prompt = BuildReadmePrompt().text()
    readme_contents = Chat(model).predict(prompt)
    FileEditor(readme_file_path).apply_changes(readme_contents)



"""
Supported LLM classes are registered here in SUPPORTED_MODELS.
"""
from nail.core.language_models.open_ai import GPT_3_5, GPT_4


SUPPORTED_MODELS = {
    "gpt-3.5-turbo": GPT_3_5,
    "gpt-4": GPT_4,
}

DEFAULT_MODEL = "gpt-3.5-turbo"



import click
from nail.tools.build_file import build_file
from nail.tools.build_readme import build_readme
from nail.tools.modify_file import modify_file
from nail.tools.debug_file import debug_file
from nail.tools.build_spec_file import build_spec_file
from nail.tools.explain_file import explain_file
from nail.core.config.user_config_utils import save_api_key

MODEL_HELP = (
    "Optionally specify an LLM model. "
    + "Currently defaults to gpt-3.5-turbo and supports gpt-4."
)


@click.group()
def main():
    pass


@main.command()
@click.option("--api_key", prompt=True, hide_input=True, help="Your OpenAI API key.")
def configure(api_key):
    save_api_key(api_key)
    click.echo("API key saved successfully.")


@main.command()
@click.argument("file")
@click.option(
    "--context-files",
    "-c",
    multiple=True,
    type=str,
    help="Optional list of context file paths.",
)
@click.option("--model", "-m", type=str, help=MODEL_HELP)
def build(file, context_files, model):
    """Build a new file with optional context files."""
    click.echo(f"Building a new file: {file}")
    if context_files:
        click.echo(f"Using context files: {', '.join(context_files)}")
    build_file(file, context_files, model)


@main.command()
@click.option("--model", "-m", type=str, help=MODEL_HELP)
def readme(model):
    """Build a new README file based on the currect directory."""
    click.echo("Generating README file.")
    build_readme("README.md", model)


@main.command()
@click.argument("file")
@click.option(
    "--request",
    "-r",
    prompt="Requested change",
    help="The modification that you are requesting.",
)
@click.option(
    "--context-files",
    "-c",
    multiple=True,
    type=str,
    help="Optional list of context file paths.",
)
@click.option("--model", "-m", type=str, help=MODEL_HELP)
def modify(file, request, context_files, model):
    """Modify an existing file."""
    click.echo(f"Modifying file: {file}")
    if context_files:
        click.echo(f"Using context files: {', '.join(context_files)}")
    modify_file(file, request, context_files, model)


@main.command()
@click.argument("file")
@click.option(
    "--error", "-e", default=None, prompt=False, help="Optional error message to debug."
)
@click.option("--model", "-m", type=str, help=MODEL_HELP)
def debug(file, error, model):
    """Debug an existing file. May include an optional error message"""
    click.echo(f"Debugging file: {file}")
    if error:
        click.echo(f"Error message: {error}")
    debug_file(file, error, model)


@main.command()
@click.argument("file")
@click.argument("target_path")
@click.option("--model", "-m", type=str, help=MODEL_HELP)
def spec(file, target_path, model):
    """Build a unit test file for an existing file."""
    click.echo(f"Building spec file for: {file}")
    click.echo(f"Target path: {target_path}")
    build_spec_file(file, target_path, model)


@main.command()
@click.argument("file")
@click.option(
    "--context-files",
    "-c",
    multiple=True,
    type=str,
    help="Optional list of context file paths.",
)
@click.option("--verbose", "-v", is_flag=True, help="Verbose output.")
@click.option("--model", "-m", type=str, help=MODEL_HELP)
def explain(file, context_files, verbose, model):
    """Explain the contents of a file."""
    click.echo(f"Explaining: {file}")
    explain_file(file, context_files, {"verbose": verbose}, model)


if __name__ == "__main__":
    main()



# Formatting Utilities


def file_block(file_path):
    """
    Returns a formatted code block containing the contents of the file at
    the given file path.

    :param file_path: The path to the file to be formatted.
    :return: A formatted code block string containing the contents of the
    file at the given file path.
    """
    with open(file_path, "r") as file:
        file_content = file.read()
    return _format_file_block(file_path, file_content)


def _format_file_block(context_file_path, context_content):
    label = _format_file_label(context_file_path)
    formatted_file_content = f"```\n{context_content}\n```"
    return f"{label}\n{formatted_file_content}\n\n"


def _format_file_label(file_path):
    return f"[[{file_path}]]"



import pytest

from unittest.mock import patch
from nail.core.file_editor import FileEditor, MissingFilePathError


def test_missing_file_path_error():
    with pytest.raises(MissingFilePathError):
        FileEditor()


def test_exists(tmp_path):
    file_path = tmp_path / "test.txt"
    file_path.write_text("Test content")

    file_editor = FileEditor(file_path)
    assert file_editor.exists() is True

    non_existent_file = tmp_path / "non_existent.txt"
    file_editor = FileEditor(non_existent_file)
    assert file_editor.exists() is False


def test_content(tmp_path):
    file_path = tmp_path / "test.txt"
    file_path.write_text("Test content")

    file_editor = FileEditor(file_path)
    assert file_editor.content() == "Test content"


@patch("subprocess.call")
def test_open_editor(mock_call, tmp_path):
    file_path = tmp_path / "test.txt"
    file_path.write_text("Test content")

    file_editor = FileEditor(file_path)
    file_editor.open_editor()

    # Check if subprocess.call was called with the correct arguments
    mock_call.assert_called_once_with(["vim", file_path])


def test_apply_changes(tmp_path, monkeypatch, capsys):
    file_path = tmp_path / "test.txt"
    file_path.write_text("Original content")

    file_editor = FileEditor(file_path)

    # Mock input to return 'y' for confirmation
    monkeypatch.setattr("builtins.input", lambda _: "y")

    assert file_editor.apply_changes("New content") is True
    assert file_editor.content() == "New content"

    # Mock input to return 'n' for discard changes
    monkeypatch.setattr("builtins.input", lambda _: "n")

    assert file_editor.apply_changes("Another content") is False
    assert file_editor.content() == "New content"

    # Check if the diff is printed correctly
    captured = capsys.readouterr()
    assert "+Another content" in captured.out
    assert "-New content" in captured.out



from rich.console import Console
from rich.markdown import Markdown
from nail.core.chat import Chat
from nail.core.prompt.prompt import ExplainPrompt


def explain_file(file_path, context_file_paths=None, verbose=False, model=None):
    prompt = ExplainPrompt(file_path, context_file_paths, {"verbose": verbose}).text()
    explanation = Chat(model).predict(prompt)
    Console().print(Markdown(explanation))



from nail.core.file_editor import FileEditor
from nail.core.chat import Chat
from nail.core.prompt.prompt import BuildPrompt


def build_file(file_path, context_file_paths=None, model=None):
    file = FileEditor(file_path)
    if not file.exists():
        file.open_editor()
    prompt = BuildPrompt(file_path, context_file_paths).text()
    draft_contents = Chat(model).predict_code(prompt)
    file.apply_changes(draft_contents)



from nail.core.chat import Chat
from nail.core.file_editor import FileEditor
from nail.core.prompt.prompt import SpecPrompt


def build_spec_file(file_path, target_file_path, model=None):
    prompt = SpecPrompt(file_path).text()
    test_file_content = Chat(model).predict_code(prompt)
    FileEditor(target_file_path).apply_changes(test_file_content)



from unittest.mock import patch, ANY
import pytest
from click.testing import CliRunner

from nail.main import configure, build, readme, modify, debug, spec


@pytest.fixture
def runner():
    return CliRunner()


def test_configure(runner):
    with patch("nail.main.save_api_key") as mock_save_api_key:
        result = runner.invoke(configure, input="test_api_key\n")
        assert result.exit_code == 0
        mock_save_api_key.assert_called_once_with("test_api_key")


def test_build(runner):
    with patch("nail.main.build_file") as mock_build_file:
        result = runner.invoke(build, ["test_file"])
        assert result.exit_code == 0
        mock_build_file.assert_called_once_with("test_file", ANY, None)


def test_readme(runner):
    with patch("nail.main.build_readme") as mock_build_readme:
        result = runner.invoke(readme)
        assert result.exit_code == 0
        mock_build_readme.assert_called_once_with("README.md", None)


def test_modify(runner):
    with patch("nail.main.modify_file") as mock_modify_file:
        result = runner.invoke(modify, ["test_file", "-r", "test_request"])
        assert result.exit_code == 0
        mock_modify_file.assert_called_once_with("test_file", "test_request", ANY, None)


def test_debug(runner):
    with patch("nail.main.debug_file") as mock_debug_file:
        result = runner.invoke(debug, ["test_file"])
        assert result.exit_code == 0
        mock_debug_file.assert_called_once_with("test_file", None, None)


def test_spec(runner):
    with patch("nail.main.build_spec_file") as mock_build_spec_file:
        result = runner.invoke(spec, ["test_file", "test_target_path"])
        assert result.exit_code == 0
        mock_build_spec_file.assert_called_once_with(
            "test_file", "test_target_path", None
        )



import re

from nail.core.loading_decorator import loadable
from nail.core.language_models.supported_models import SUPPORTED_MODELS, DEFAULT_MODEL


class InvalidModelError(Exception):
    pass


class ModelCodeGenerationError(Exception):
    pass


class Chat:
    def __init__(self, model_name=None):
        self._set_llm_model(model_name)

    def predict(self, prompt):
        create_completion = loadable(self.model.respond, "Loading response...")
        return create_completion(prompt)

    def predict_code(self, prompt):
        create_completion = loadable(self.model.respond_with_code, "Generating code...")
        completion = create_completion(prompt)
        return self._extract_code(completion)

    def _extract_code(self, text):
        # regex for text wrapped in triple backticks
        # and optional language identifier
        code_block_pattern = r"^```(?:\w+)?\n([\s\S]*?)\n```"
        code_blocks = re.findall(code_block_pattern, text, re.MULTILINE)
        try:
            return code_blocks[0]
        except IndexError:
            err_msg = f"Model ({self.model_name}) failed to respond with code"
            raise ModelCodeGenerationError(err_msg)

    def _set_llm_model(self, model_name):
        self.model_name = DEFAULT_MODEL if model_name is None else model_name
        try:
            self.model = SUPPORTED_MODELS[self.model_name](self.model_name)
        except KeyError:
            raise InvalidModelError(f"Unsupported model: {model_name}")



from nail.core.file_editor import FileEditor
from nail.core.chat import Chat, ModelCodeGenerationError
from nail.core.prompt.prompt import DebugPrompt

NO_BUGS_MESSAGE = "No bugs were found in the file."


def debug_file(file_path, error_message, model=None):
    file = FileEditor(file_path)
    prompt = DebugPrompt(file_path, details={"error_message": error_message}).text()
    try:
        modified_contents = Chat(model).predict_code(prompt)
    except ModelCodeGenerationError:
        print(NO_BUGS_MESSAGE)
        return
    file.apply_changes(modified_contents)



import pytest

from unittest.mock import MagicMock, patch

from nail.core.language_models.supported_models import SUPPORTED_MODELS, DEFAULT_MODEL
from nail.core.chat import Chat, InvalidModelError, ModelCodeGenerationError


def test_init_with_default_model():
    chat = Chat()
    assert chat.model_name == DEFAULT_MODEL
    assert type(chat.model) == SUPPORTED_MODELS[DEFAULT_MODEL]


def test_init_with_custom_model():
    custom_model_name = "custom_model"
    SUPPORTED_MODELS[custom_model_name] = MagicMock(
        return_value="custom_model_instance"
    )

    chat = Chat(custom_model_name)
    assert chat.model_name == custom_model_name
    assert chat.model == "custom_model_instance"

    del SUPPORTED_MODELS[custom_model_name]


def test_init_with_invalid_model():
    with pytest.raises(InvalidModelError):
        Chat("invalid_model")


def test_predict():
    chat = Chat()
    chat.model.respond = MagicMock(return_value="response")
    prompt = "test_prompt"

    with patch(
        "nail.core.loading_decorator.loadable",
        MagicMock(return_value=chat.model.respond),
    ):
        response = chat.predict(prompt)

    chat.model.respond.assert_called_once_with(prompt)
    assert response == "response"


def test_predict_code():
    chat = Chat()
    chat.model.respond_with_code = MagicMock(return_value="```\ncode\n```")
    prompt = "test_prompt"

    with patch(
        "nail.core.loading_decorator.loadable",
        MagicMock(return_value=chat.model.respond_with_code),
    ):
        code = chat.predict_code(prompt)

    chat.model.respond_with_code.assert_called_once_with(prompt)
    assert code == "code"


def test_predict_code_with_invalid_response():
    chat = Chat()
    chat.model.respond_with_code = MagicMock(return_value="no code block")
    prompt = "test_prompt"

    with patch(
        "nail.core.loading_decorator.loadable",
        MagicMock(return_value=chat.model.respond_with_code),
    ):
        with pytest.raises(ModelCodeGenerationError):
            chat.predict_code(prompt)



import configparser
from unittest import mock
from unittest.mock import patch
import pytest
from nail.core.config.user_config_utils import get_api_key, save_api_key


@pytest.fixture
def temp_config_file(tmp_path):
    temp_file = tmp_path / ".nailrc"
    temp_file.touch()
    return temp_file


def test_get_api_key_with_env_var(monkeypatch):
    monkeypatch.setenv("OPENAI_API_KEY", "test_key")
    assert get_api_key() == "test_key"


def test_get_api_key_with_config_file(monkeypatch, temp_config_file):
    monkeypatch.delenv("OPENAI_API_KEY", raising=False)
    config = configparser.ConfigParser()
    config["openai"] = {"api_key": "test_key"}
    with temp_config_file.open("w") as f:
        config.write(f)

    with mock.patch(
        "nail.core.config.user_config_utils.USER_CONFIG_FILE", temp_config_file
    ):
        assert get_api_key() == "test_key"


def test_get_api_key_with_no_key(monkeypatch, temp_config_file):
    monkeypatch.delenv("OPENAI_API_KEY", raising=False)
    with mock.patch(
        "nail.core.config.user_config_utils.USER_CONFIG_FILE", temp_config_file
    ):
        assert get_api_key() is None


def test_save_api_key(tmp_path):
    api_key = "test_api_key"
    temp_config_file = tmp_path / ".nailrc"

    with patch("nail.core.config.user_config_utils.USER_CONFIG_FILE", temp_config_file):
        save_api_key(api_key)
        config = configparser.ConfigParser()
        config.read(temp_config_file)
        assert config["openai"]["api_key"] == api_key



from nail.core.file_editor import FileEditor
from nail.core.chat import Chat
from nail.core.prompt.prompt import ModifyPrompt


def modify_file(file_path, request, context_file_paths=None, model=None):
    prompt = ModifyPrompt(
        file_path, context_file_paths, details={"request": request}
    ).text()
    modified_contents = Chat(model).predict_code(prompt)
    FileEditor(file_path).apply_changes(modified_contents)



import os
import yaml

CONFIG_FILE_NAME = ".nail.yaml"


def load_local_config():
    config = {}
    if os.path.exists(CONFIG_FILE_NAME):
        with open(CONFIG_FILE_NAME, "r") as file:
            config = yaml.safe_load(file)
    return config



from abc import ABC, abstractmethod

from nail.core.file_editor import FileEditor
from nail.core.prompt.context_compiler import ContextCompiler
from nail.core.prompt.formatting_utils import file_block
from nail.core.config.local_config_utils import load_local_config

BUILD_REQUEST = "Write code to the following specification:"
ERROR_REQUEST = "Fix the following error message:"
EXPLAIN_PREFIX = "Explain the following code:"
GENERAL_DEBUG_REQUEST = "Fix any bugs in the file."
ORIGINAL_FILE_TAG = "Original file contents:"
README_REQUEST = "Generate a README file for the application."
REQUEST_TAG = "Request:"
RETURN_FULL_FILE = (
    "Return the full modified file contents. Any non-code"
    + " text should only be included as inline comments."
)
SPEC_PREFIX = "Create a unit test file for the following code:"
LOW_VERBOSITY = "Keep your answer succinct and to the point."
HIGH_VERBOSITY = "Include a high level of detail."


class BasePrompt(ABC):
    """
    Base class for all prompts. Contains common functionality for all prompts.
    All prompts should implement the text method.

    :param file_path: Path to the main file to be edited, debugged, etc
    :param context_file_paths: Paths to other relevant files for context
    :param details: A dictionary of details to be used by specific prompts
    """

    def __init__(self, file_path=None, context_file_paths=[], details={}):
        self.file_path = file_path
        self.context_file_paths = context_file_paths
        self.details = details

    @property
    def _context_text(self):
        if not self.context_file_paths:
            return ""
        return ContextCompiler(self.context_file_paths).compile_all()

    @property
    def _file_text(self):
        return FileEditor(self.file_path).content()

    def _custom_instructions(self, key):
        instruction = load_local_config().get("prompt_instructions", {}).get(key)
        return "" if not instruction else f"\n{instruction}"

    @abstractmethod
    def text(self):
        pass


class BuildPrompt(BasePrompt):
    def text(self):
        return (
            self._context_text
            + f"{BUILD_REQUEST}\n"
            + self._file_text
            + self._custom_instructions("build")
        )


class BuildReadmePrompt(BasePrompt):
    def text(self):
        return self._context_text + README_REQUEST + self._custom_instructions("readme")

    @property
    def _context_text(self):
        return ContextCompiler().compile_all_minus_ignored()


class DebugPrompt(BasePrompt):
    def text(self):
        return (
            file_block(self.file_path)
            + f"{self._debug_request}\n"
            + RETURN_FULL_FILE
            + self._custom_instructions("debug")
        )

    @property
    def _debug_request(self):
        error_message = self.details.get("error_message")
        if error_message:
            return f"{ERROR_REQUEST}\n{error_message}"
        else:
            return GENERAL_DEBUG_REQUEST


class ModifyPrompt(BasePrompt):
    def text(self):
        file_context = f"{ORIGINAL_FILE_TAG}\n{file_block(self.file_path)}"
        return (
            self._context_text
            + file_context
            + self._modify_request
            + self._custom_instructions("modify")
        )

    @property
    def _modify_request(self):
        request = self.details.get("request")
        return f"{REQUEST_TAG} {request}\n{RETURN_FULL_FILE}"


class SpecPrompt(BasePrompt):
    def text(self):
        return (
            f"{SPEC_PREFIX}\n{file_block(self.file_path)}"
            + self._custom_instructions("spec")
        )


class ExplainPrompt(BasePrompt):
    def text(self):
        if self.details.get("verbose") is True:
            verbosity = HIGH_VERBOSITY
        else:
            verbosity = LOW_VERBOSITY
        return (
            self._context_text
            + f"{EXPLAIN_PREFIX}\n{file_block(self.file_path)}"
            + verbosity
            + self._custom_instructions("explain")
        )



from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="nail",
    version="0.1.2",
    description="A CLI tool for speeding up development using LLMs",
    packages=find_packages(),
    install_requires=[
        "click",
        "openai",
        "termcolor",
        "PyYAML",
        "rich",
    ],
    entry_points="""
        [console_scripts]
        nail=nail.main:main
    """,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Edward Saavedra",
    author_email="edsaav@gmail.com",
    url="https://github.com/edsaav/nail",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import openai

from nail.core.config.user_config_utils import get_api_key
from nail.core.language_models.language_model import LanguageModel


DEFAULT_TEMPERATURE = 0.3
DEFAULT_MAX_TOKENS = 2048


class OpenAIAPIError(Exception):
    pass


class OpenAIChat(LanguageModel):
    """
    Implementation for the OpenAI chat (*not completion*) API.
    Reference: https://platform.openai.com/docs/guides/chat/introduction
    """

    def _user_message(self, message):
        return {"role": "user", "content": message}

    def _fetch_response(self, messages):
        openai.api_key = get_api_key()
        response = openai.ChatCompletion.create(
            model=self.model_name,
            messages=messages,
            temperature=DEFAULT_TEMPERATURE,
            max_tokens=DEFAULT_MAX_TOKENS,
        )
        return response

    def _parse_response(self, response):
        try:
            return response.choices[0].message["content"]
        except (AttributeError, IndexError):
            raise OpenAIAPIError("OpenAI API response is invalid.")


class GPT_3_5(OpenAIChat):
    """
    This implementation instead prefixes the initial user message with the
    system message content. This should result in more accurate instruction
    following.

    From the OpenAI docs:
    "gpt-3.5-turbo-0301 does not always pay strong attention to system messages.
    Future models will be trained to pay stronger attention to system messages."
    """

    def _respond_from_prompt(self, prompt, system_message):
        full_prompt = f"{system_message}\n{prompt}"
        messages = [self._user_message(full_prompt)]
        return self._parse_response(self._fetch_response(messages))


class GPT_4(OpenAIChat):
    """
    The GPT-4 model makes full use of the concept of system messages, so
    this implementation uses them as intended.
    """

    def _system_message(self, message):
        return {"role": "system", "content": message}

    def _respond_from_prompt(self, prompt, system_message):
        messages = [self._system_message(system_message), self._user_message(prompt)]
        return self._parse_response(self._fetch_response(messages))
